@@ -11,14 +11,14 @@ import openturns as ot
 import numpy as np
 import os
 
-sample_E = ot.NumericalSample.ImportFromCSVFile(
+sample_E = ot.Sample.ImportFromCSVFile(
     os.path.join(os.path.dirname(__file__), "sample_E.csv"))
 kernel_smoothing = ot.KernelSmoothing(ot.Normal())
 bandwidth = kernel_smoothing.computeSilvermanBandwidth(sample_E)
 E = kernel_smoothing.build(sample_E, bandwidth)
 E.setDescription(['Young modulus'])
 
-F = ot.LogNormal(30000, 9000, 15000, ot.LogNormal.MUSIGMA)
+F = ot.LogNormalMuSigma(30000.0, 9000.0, 15000.0).getDistribution()
 F.setDescription(['Load'])
 
 L = ot.Uniform(250, 260)
@@ -33,4 +33,4 @@ SR_cor = ot.CorrelationMatrix(len(marginal_distributions))
 SR_cor[2, 3] = -0.2
 copula = ot.NormalCopula(ot.NormalCopula.GetCorrelationFromSpearmanCorrelation(SR_cor))
 
-X_distribution = ot.ComposedDistribution(marginal_distributions, copula)
\ No newline at end of file
+X_distribution = ot.ComposedDistribution(marginal_distributions, copula)
